@@ -1,60 +1,72 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
-
+import logging
 from collections import defaultdict
 from typing import Any, DefaultDict, Dict, List, Set, Union
 
-from arango import ArangoClient
 from arango.cursor import Cursor
-from arango.graph import Graph as ArangoDBGraph
+from arango.database import Database
+from arango.graph import Graph as ADBGraph
 from arango.result import Result
-from dgl import DGLGraph, heterograph
-from dgl.heterograph import DGLHeteroGraph
+from dgl import DGLGraph, DGLHeteroGraph, heterograph
 from dgl.view import HeteroEdgeDataView, HeteroNodeDataView
-from torch import tensor  # type: ignore
+from torch import tensor
 from torch.functional import Tensor
 
 from .abc import Abstract_ADBDGL_Adapter
 from .controller import ADBDGL_Controller
 from .typings import ArangoMetagraph, DGLCanonicalEType, DGLDataDict, Json
+from .utils import logger
 
 
 class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
     """ArangoDB-DGL adapter.
 
-    :param conn: Connection details to an ArangoDB instance.
-    :type conn: adbdgl_adapter.typings.Json
+    :param db: A python-arango database instance
+    :type db: arango.database.Database
     :param controller: The ArangoDB-DGL controller, for controlling how
         ArangoDB attributes are converted into DGL features, and vice-versa.
         Optionally re-defined by the user if needed (otherwise defaults to
         ADBDGL_Controller).
     :type controller: adbdgl_adapter.controller.ADBDGL_Controller
-    :raise ValueError: If missing required keys in conn
+    :param logging_lvl: Defaults to logging.INFO. Other useful options are
+        logging.DEBUG (more verbose), and logging.WARNING (less verbose).
+    :type logging_lvl: str | int
+    :raise ValueError: If invalid parameters
     """
 
     def __init__(
         self,
-        conn: Json,
+        db: Database,
         controller: ADBDGL_Controller = ADBDGL_Controller(),
+        logging_lvl: Union[str, int] = logging.INFO,
     ):
-        self.__validate_attributes("connection", set(conn), self.CONNECTION_ATRIBS)
-        if issubclass(type(controller), ADBDGL_Controller) is False:
-            msg = "controller must inherit from ADBDGL_Controller"
+        self.set_logging(logging_lvl)
+
+        if issubclass(type(db), Database) is False:
+            msg = "**db** parameter must inherit from arango.database.Database"
             raise TypeError(msg)
 
-        username: str = conn["username"]
-        password: str = conn["password"]
-        db_name: str = conn["dbName"]
-        host: str = conn["hostname"]
-        protocol: str = conn.get("protocol", "https")
-        port = str(conn.get("port", 8529))
+        if issubclass(type(controller), ADBDGL_Controller) is False:
+            msg = "**controller** parameter must inherit from ADBDGL_Controller"
+            raise TypeError(msg)
 
-        url = protocol + "://" + host + ":" + port
-
-        print(f"Connecting to {url}")
-        self.__db = ArangoClient(hosts=url).db(db_name, username, password, verify=True)
+        self.__db = db
         self.__cntrl: ADBDGL_Controller = controller
 
+        logger.info(f"Instantiated ADBDGL_Adapter with database '{db.name}'")
+
+    @property
+    def db(self) -> Database:
+        return self.__db  # pragma: no cover
+
+    @property
+    def cntrl(self) -> ADBDGL_Controller:
+        return self.__cntrl  # pragma: no cover
+
+    def set_logging(self, level: Union[int, str]) -> None:
+        logger.setLevel(level)
+
     def arangodb_to_dgl(
         self, name: str, metagraph: ArangoMetagraph, **query_options: Any
     ) -> DGLHeteroGraph:
@@ -89,6 +101,7 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
             },
         }
         """
+        logger.debug(f"Starting arangodb_to_dgl({name}, ...):")
         self.__validate_attributes("graph", set(metagraph), self.METAGRAPH_ATRIBS)
 
         # Maps ArangoDB vertex IDs to DGL node IDs
@@ -101,6 +114,7 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
 
         adb_v: Json
         for v_col, atribs in metagraph["vertexCollections"].items():
+            logger.debug(f"Preparing '{v_col}' vertices")
             for i, adb_v in enumerate(
                 self.__fetch_adb_docs(v_col, atribs, query_options)
             ):
@@ -115,6 +129,7 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
         from_col: Set[str] = set()
         to_col: Set[str] = set()
         for e_col, atribs in metagraph["edgeCollections"].items():
+            logger.debug(f"Preparing '{e_col}' edges")
             from_nodes: List[int] = []
             to_nodes: List[int] = []
             for adb_e in self.__fetch_adb_docs(e_col, atribs, query_options):
@@ -143,11 +158,12 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
         dgl_g: DGLHeteroGraph = heterograph(data_dict)
         has_one_ntype = len(dgl_g.ntypes) == 1
         has_one_etype = len(dgl_g.etypes) == 1
+        logger.debug(f"Is graph '{name}' homogenous? {has_one_ntype and has_one_etype}")
 
         self.__insert_dgl_features(ndata, dgl_g.ndata, has_one_ntype)
         self.__insert_dgl_features(edata, dgl_g.edata, has_one_etype)
 
-        print(f"DGL: {name} created")
+        logger.info(f"Created DGL '{name}' Graph")
         return dgl_g
 
     def arangodb_collections_to_dgl(
@@ -198,7 +214,7 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
 
     def dgl_to_arangodb(
         self, name: str, dgl_g: Union[DGLGraph, DGLHeteroGraph], batch_size: int = 1000
-    ) -> ArangoDBGraph:
+    ) -> ADBGraph:
         """Create an ArangoDB graph from a DGL graph.
 
         :param name: The ArangoDB graph name.
@@ -210,32 +226,33 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
         :return: The ArangoDB Graph API wrapper.
         :rtype: arango.graph.Graph
         """
+        logger.debug(f"Starting dgl_to_arangodb({name}, ...):")
+
         is_default = dgl_g.canonical_etypes == self.DEFAULT_CANONICAL_ETYPE
-        adb_v_cols: List[str] = [name + dgl_g.ntypes[0]] if is_default else dgl_g.ntypes
-        adb_e_cols: List[str] = [name + dgl_g.etypes[0]] if is_default else dgl_g.etypes
-        e_definitions = self.etypes_to_edefinitions(
-            [
-                (
-                    adb_v_cols[0],
-                    adb_e_cols[0],
-                    adb_v_cols[0],
-                )
-            ]
+        logger.debug(f"Is graph '{name}' using default canonical_etypes? {is_default}")
+
+        edge_definitions = self.etypes_to_edefinitions(
+            [(name + "_N", name + "_E", name + "_N")]
             if is_default
             else dgl_g.canonical_etypes
         )
 
-        has_one_ntype = len(dgl_g.ntypes) == 1
-        has_one_etype = len(dgl_g.etypes) == 1
+        self.__db.delete_graph(name, ignore_missing=True)
+        adb_graph: ADBGraph = self.__db.create_graph(name, edge_definitions)
+
+        adb_v_cols = adb_graph.vertex_collections()
+        adb_e_cols = [e_d["edge_collection"] for e_d in edge_definitions]
+
+        has_one_vcol = len(adb_v_cols) == 1
+        has_one_ecol = len(adb_e_cols) == 1
+        logger.debug(f"Is graph '{name}' homogenous? {has_one_vcol and has_one_ecol}")
 
         adb_documents: DefaultDict[str, List[Json]] = defaultdict(list)
         for v_col in adb_v_cols:
             ntype = None if is_default else v_col
             v_col_docs = adb_documents[v_col]
 
-            if self.__db.has_collection(v_col) is False:
-                self.__db.create_collection(v_col)
-
+            logger.debug(f"Preparing {len(dgl_g.nodes(ntype))} '{v_col}' DGL nodes")
             node: Tensor
             for node in dgl_g.nodes(ntype):
                 dgl_node_id = node.item()
@@ -246,7 +263,7 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
                     dgl_node_id,
                     adb_vertex,
                     v_col,
-                    has_one_ntype,
+                    has_one_vcol,
                 )
 
                 self.__insert_adb_docs(v_col, v_col_docs, adb_vertex, batch_size)
@@ -259,15 +276,13 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
             etype = None if is_default else e_col
             e_col_docs = adb_documents[e_col]
 
-            if self.__db.has_collection(e_col) is False:
-                self.__db.create_collection(e_col, edge=True)
-
             if is_default:
                 from_col = to_col = adb_v_cols[0]
             else:
                 from_col, _, to_col = dgl_g.to_canonical_etype(e_col)
 
             from_nodes, to_nodes = dgl_g.edges(etype=etype)
+            logger.debug(f"Preparing {len(from_nodes)} '{e_col}' DGL edges")
             for dgl_edge_id, (from_node, to_node) in enumerate(
                 zip(from_nodes, to_nodes)
             ):
@@ -282,18 +297,17 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
                     dgl_edge_id,
                     adb_edge,
                     e_col,
-                    has_one_etype,
+                    has_one_ecol,
                 )
 
                 self.__insert_adb_docs(e_col, e_col_docs, adb_edge, batch_size)
 
-        self.__db.delete_graph(name, ignore_missing=True)
-        adb_graph: ArangoDBGraph = self.__db.create_graph(name, e_definitions)
-
         for col, doc_list in adb_documents.items():  # insert remaining documents
-            self.__db.collection(col).import_bulk(doc_list, on_duplicate="replace")
+            if doc_list:
+                logger.debug(f"Inserting last {len(doc_list)} documents into '{col}'")
+                self.__db.collection(col).import_bulk(doc_list, on_duplicate="replace")
 
-        print(f"ArangoDB: {name} created")
+        logger.info(f"Created ArangoDB '{name}' Graph")
         return adb_graph
 
     def etypes_to_edefinitions(
@@ -375,6 +389,7 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
         col_dict: Dict[str, List[Any]]
         for key, col_dict in features_data.items():
             for col, array in col_dict.items():
+                logger.debug(f"Inserting {len(array)} '{key}' features into '{col}'")
                 data[key] = (
                     tensor(array) if has_one_type else {**data[key], col: tensor(array)}
                 )
@@ -386,7 +401,7 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
         id: Union[int, float, bool],
         doc: Json,
         col: str,
-        has_one_type: bool,
+        has_one_col: bool,
     ) -> None:
         """Convert DGL features into a set of ArangoDB attributes for a given document
 
@@ -401,12 +416,12 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
         :type doc: adbdgl_adapter.typings.Json
         :param col: The collection the current document belongs to
         :type col: str
-        :param has_one_type: Set to True if the DGL graph only has one ntype,
-            or one etype.
-        :type has_one_type: bool
+        :param has_one_col: Set to True if the ArangoDB graph has one
+            vertex collection or one edge collection only.
+        :type has_one_col: bool
         """
         for key in features:
-            tensor = data[key] if has_one_type else data[key][col]
+            tensor = data[key] if has_one_col else data[key][col]
             doc[key] = self.__cntrl._dgl_feature_to_adb_attribute(key, col, tensor[id])
 
     def __insert_adb_docs(
@@ -431,6 +446,7 @@ class ADBDGL_Adapter(Abstract_ADBDGL_Adapter):
         col_docs.append(doc)
 
         if len(col_docs) >= batch_size:
+            logger.debug(f"Inserting next {batch_size} batch documents into '{col}'")
             self.__db.collection(col).import_bulk(col_docs, on_duplicate="replace")
             col_docs.clear()
 
